open Ast
open Lazy
open Format


(** prune l is the list of (a, l') where the elements of l and a::l' are
 * exactly the same. *)
let prune l =
  let rec aux before l' = function
    | [] -> l'
    | a::l -> aux (a::before) ((a, before $ l)::l') l in
  aux [] [] l


type 'a stream = Wait of 'a * 'a stream lazy_t

(** variables is the stream: x, y, z, u, v, w, t, x1, y1, ... *)
let variables =
  let alphabet = ["t";"w";"v";"u";"z";"y";"x"] in
  let (a,l) = match alphabet with
    | []   -> ("",[])
    | a::l -> (a,l) in
  let rec f n =
    let m = string_of_int n in
    List.fold_left (fun w x -> Wait (x^"_"^m, lazy w))
      (Wait (a^"_"^m, lazy (f (n+1)))) l in
  if a = "" then f 0 else
  List.fold_left (fun w x -> Wait (x, lazy w)) (Wait (a, lazy (f 1))) l

(** canonical is the stream: _0, _1, _2, _3, ... *)
let canonical =
  let rec f n =
    let m = string_of_int n in
    Wait ("_"^m, lazy (f (n+1))) in
  f 0


let empty = SMap.empty

(** shrink is the union of two channel maps. *)
let shrink m =
  SMap.union (fun _ (s,r) (s',r') -> Some (s' $ s, r' $ r)) m

(** merge is the union of two programs. *)
let merge (Prog (m,l)) (Prog (m',l')) =
  Prog (shrink m m', List.merge Pervasives.compare l l')


(** assemble needs the list to be sorted. *)
let rec assemble_aux action l' env current = function
  | [] -> l'
  | (x::l as v) -> (match env with
      | [] -> assert false
      | a::e -> if current < x then assemble_aux action l' e (current + 1) v
        else assemble_aux action ((action a)::l') env current l)
let assemble action env l =
  assemble_aux action [] env 0 l

let rec instantiate env (Prog (m, l): prog) ((Wait (x, f)) as w) =
  let rec inst_send_list l' = function
    | []   -> (l': disp_send list)
    | s::l -> inst_send_list ((instantiate env s w)::l') l in
  let rec inst_rec_list l' = function
    | []   -> (l': disp_receive list)
    | r::l -> inst_rec_list ((x, instantiate (x::env) r (force f))::l') l in
  let inst_rec_var = assemble (fun x -> x) in
  (DProg (channel_map (fun (s,r) -> (inst_send_list [] s, inst_rec_list [] r)) m,
          inst_rec_var env l): disp_prog)

(** surround transforms an inprog into a toplevel by adding the same environment
 * to each subprogram *)
let surround env neutral m =
  let rec add_end l' = function
    | []   -> l'
    | a::l -> add_end ((a,(if a = Prog(SMap.empty,[]) then neutral
                           else env))::l') l in
  SMap.map (fun (s,r) -> (add_end [] s, add_end [] r)) m

(** eval p e evaluates program p in environment e, making it a toplevel *)
let rec eval (Prog (m, l)) ((Env e) as env) =
  let action = assemble (fun (p,e) -> eval p e) in
  let x = surround env (Env []) m in
  (List.fold_left shrink SMap.empty (x::(action e l)): toplevel)


(** remove_first x l removes the first occurence of x in l
 * The returned list may not have the same order as the original. *)
let remove_first x l =
  let rec aux l' = function
    | []   -> raise Not_found
    | a::l -> if a = x then l' $ l else aux (a::l') l in
  aux [] l

(** compare_list compares two list term by term *)
let rec compare_list compare = function
  | (a::l1), (b::l2) -> let x = compare a b in
    if x <> 0 then x else compare_list compare (l1,l2)
  | [], [] -> 0
  | [], _  -> -1
  | _      -> 1


(** sort_compare compares the contents of two lists, regardless of the initial
 * order *)
let sort_compare compare l1 l2 =
  if l1 == l2 then 0 else
    let l1' = List.sort compare l1 in
    let l2' = List.sort compare l2 in
    compare_list compare (l1', l2')

(** compare_inprog is a comparison function for inprogs *)
let rec compare_inprog m1 m2 =
  if m1 == m2 then 0 else
    SMap.compare (fun (s1,r1) (s2,r2) ->
        let y = sort_compare compare_prog s1 s2 in
        if y <> 0 then y else
          sort_compare compare_prog r1 r2) m1 m2

(** compare_prog is a comparison function for programs, suitable for Set *)
and compare_prog (Prog (m1, l1) as p1) (Prog (m2, l2) as p2) =
  if p1 == p2 then 0 else
    let x = compare_list Pervasives.compare (l1,l2) in
    if x <> 0 then x else
    compare_inprog m1 m2

module PSet = Set.Make(struct
    type t = prog
    let compare = compare_prog end)

(** compare_env is a comparison function for environments *)
let rec compare_env (Env e1 as env1) (Env e2 as env2) =
  if env1 == env2 then 0 else
    compare_list compare_annot (e1,e2)

(** compare_annot is a comparison function for annotated programs *)
and compare_annot ((p1,e1) as a1) ((p2,e2) as a2) =
  if a1 == a2 then 0 else
    let x = compare_prog p1 p2 in
    if x <> 0 then x else
      compare_env e1 e2

(** compare_top is a comparison function for toplevels *)
let compare_top (m1: toplevel) (m2: toplevel) =
  if m1 == m2 then 0 else
    SMap.compare (fun (s1, r1) (s2, r2) ->
        let y = sort_compare compare_annot s1 s2 in
        if y <> 0 then y else
          sort_compare compare_annot r1 r2) m1 m2

module TSet = Set.Make(struct
    type t = toplevel
    let compare = compare_top end)

